@@ -6,5 +6,8 @@ use Java::JVM::Classfile::Perl;
 
 my $c = Java::JVM::Classfile::Perl->new(shift || "HelloWorld.class");
 
-print $c->as_perl();
-eval $c->as_perl();
+my $perl = $c->as_perl;
+#print $perl;
+print eval $perl;
+
+
